﻿using Microsoft.DwayneNeed.Utilities;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.WebSockets;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Web.Script.Serialization;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using Microsoft.DwayneNeed.Numerics;
using Microsoft.DwayneNeed.Win32.User32;
using Microsoft.DwayneNeed.Win32;

namespace RcpServer
{
    public class Connection
    {
        public Connection(WebSocket webSocket, ScreenScraperBitmapBuffer screenScraper)
        {
            _ackRandom = new Random();

            _serializer = new JavaScriptSerializer();
            _serializer.MaxJsonLength = 10000000; // 10 million chars

            _webSocket = webSocket;
            _cancellationToken = new CancellationTokenSource();

            _screenScraper = screenScraper;
            _serverBuffer = new DirtyRectBitmapBuffer(screenScraper.Width, screenScraper.Height, MAX_DIRTYRECTS);
            _serverBuffer.SetBits(_screenScraper.View, new Int32Rect(0, 0, screenScraper.Width, screenScraper.Height));

            _screenScraper.Resized += ScreenScraper_Resized;
            _screenScraper.Updated += ScreenScraper_Updated;

            // Send the client the first message.
            Task t = SendNextClientMessage();
            t.Wait();
        }

        public async void StartReader()
        {
            byte[] readBuffer = new byte[1024];
            ArraySegment<byte> readBufferSegment = new ArraySegment<byte>(readBuffer);

            try
            {
                while (!_cancellationToken.IsCancellationRequested && _webSocket != null)
                {
                    WebSocketReceiveResult result = await _webSocket.ReceiveAsync(readBufferSegment, _cancellationToken.Token);

                    if (result.MessageType != WebSocketMessageType.Text)
                    {
                        // We only support text messages.
                        Close(WebSocketCloseStatus.InvalidMessageType, "only text messages are supported");
                        break;
                    }

                    if (!result.EndOfMessage)
                    {
                        // The message was longer than our buffer?  Or perhapse it was spit at the client side?
                        Close(WebSocketCloseStatus.MessageTooBig, "message was not complete");
                        break;
                    }


                    string message = Encoding.ASCII.GetString(readBuffer, 0, result.Count);
                    Dictionary<string, object> packet = (Dictionary<string, object>)_serializer.DeserializeObject(message);

                    string type = (string)packet["Type"];
                    if (type == "Ack")
                    {
                        int messageId = (int)packet["MessageId"];

                        int expectedMessageId = _pendingAckQueue.Dequeue();
                        if (messageId != expectedMessageId)
                        {
                            Close(WebSocketCloseStatus.PolicyViolation, "incorrect ACK");
                        }

                        if (!_sending && _pendingAckQueue.Count < MAX_PENDING_ACK)
                        {
                            await SendNextClientMessage();
                        }
                    }
                    else if (type == "MouseMove")
                    {
                        int x = ToInt(packet["X"]);
                        int y = ToInt(packet["Y"]);

                        // The coordinates passed to SendInput are normalized
                        // to be bewtween 0 and 65535, inclusive.
                        //
                        // Note: multimon would require special handling.  See
                        // MOUSEEVENTF.VIRTUALDESK.
                        x = (x * 65535) / _clientBuffer.Width;
                        y = (y * 65535) / _clientBuffer.Height;

                        INPUT[] inputs = new INPUT[1];
                        inputs[0].type = INPUTTYPE.MOUSE;
                        inputs[0].mouseInput.dx = Decimal.ToInt32(x);
                        inputs[0].mouseInput.dy = Decimal.ToInt32(y);
                        inputs[0].mouseInput.dwFlags = MOUSEEVENTF.ABSOLUTE | MOUSEEVENTF.MOVE;

                        NativeMethods.SendInput(1, inputs, Marshal.SizeOf(inputs[0]));
                    }
                    else if (type == "MouseDown")
                    {
                        int button = ToInt(packet["Button"]);

                        INPUT[] inputs = new INPUT[1];
                        inputs[0].type = INPUTTYPE.MOUSE;
                        switch (button)
                        {
                            case 0: inputs[0].mouseInput.dwFlags = MOUSEEVENTF.LEFTDOWN; break;
                            case 1: inputs[0].mouseInput.dwFlags = MOUSEEVENTF.MIDDLEDOWN; break;
                            case 2: inputs[0].mouseInput.dwFlags = MOUSEEVENTF.RIGHTDOWN; break;
                        }

                        NativeMethods.SendInput(1, inputs, Marshal.SizeOf(inputs[0]));
                    }
                    else if (type == "MouseUp")
                    {
                        int button = ToInt(packet["Button"]);

                        INPUT[] inputs = new INPUT[1];
                        inputs[0].type = INPUTTYPE.MOUSE;
                        switch (button)
                        {
                            case 0: inputs[0].mouseInput.dwFlags = MOUSEEVENTF.LEFTUP; break;
                            case 1: inputs[0].mouseInput.dwFlags = MOUSEEVENTF.MIDDLEUP; break;
                            case 2: inputs[0].mouseInput.dwFlags = MOUSEEVENTF.RIGHTUP; break;
                        }

                        NativeMethods.SendInput(1, inputs, Marshal.SizeOf(inputs[0]));
                    }
                    else if (type == "KeyDown")
                    {
                        short keyCode = (short)ToInt(packet["KeyCode"]);

                        INPUT[] inputs = new INPUT[1];
                        inputs[0].type = INPUTTYPE.KEYBOARD;
                        inputs[0].keyboardInput.wVk = keyCode;
                        inputs[0].keyboardInput.dwFlags = 0;

                        NativeMethods.SendInput(1, inputs, Marshal.SizeOf(inputs[0]));
                    }
                    else if (type == "KeyUp")
                    {
                        short keyCode = (short)ToInt(packet["KeyCode"]);

                        INPUT[] inputs = new INPUT[1];
                        inputs[0].type = INPUTTYPE.KEYBOARD;
                        inputs[0].keyboardInput.wVk = keyCode;
                        inputs[0].keyboardInput.dwFlags = KEYEVENTF.KEYUP;

                        NativeMethods.SendInput(1, inputs, Marshal.SizeOf(inputs[0]));
                    }
                    else if (type == "QOS")
                    {
                        string format = (string)packet["Format"];
                        int qualityLevel = packet.ContainsKey("QualityLevel") ? ToInt(packet["QualityLevel"]) : 0;

                        if (format == "png")
                        {
                            _encoderType = 0;
                            _qualityLevel = qualityLevel; // 0 == normal, 1 == XOR
                        }
                        else if (format == "jpg")
                        {
                            _encoderType = 1;
                            _qualityLevel = qualityLevel;
                        }
                    }
                }
            }
            catch (Exception)
            {
                Close(WebSocketCloseStatus.InternalServerError, "error");
            }
        }

        public async void Close(WebSocketCloseStatus status, string statusDescription)
        {
            if (_webSocket != null)
            {
                WebSocket webSocket = _webSocket;

                _cancellationToken.Cancel();

                // TODO: calling CloseAsync sometimes results in an exception
                // because the reader is still reading messages, and when one
                // arrives it barfs because the socket is in a closed state.
                // For now, just use CloseOutputAsync.
                Console.WriteLine("Closing connection.");
                await webSocket.CloseOutputAsync(status, statusDescription, CancellationToken.None);
            }
            else
            {
                throw new InvalidOperationException("WebSocketConnection already closed.");
            }
        }

        private async Task SendNextClientMessage()
        {
            if (_sending == true)
            {
                Console.WriteLine("***** Critical Error: SendNextClientMessage called while _sending is already true.");
            }

            bool actionTaken = true;
            _sending = true;

            try
            {
                do
                {
                    if (_clientPendingBuffer == null)
                    {
                        // This is the first time we have sent a message to
                        // the client, so the client does not have any of the
                        // contents yet.  We will need to transfer the entire
                        // contents the server buffer, which we will send as
                        // one dirty rect.
                        _clientPendingBuffer = new DirtyRectBitmapBuffer(_serverBuffer.Width, _serverBuffer.Height, MAX_DIRTYRECTS);
                        _clientBuffer = new WriteableBitmapBuffer(_serverBuffer.Width, _serverBuffer.Height);
                        _clientPendingBuffer.SetBits(_serverBuffer.View, new Int32Rect(0, 0, _serverBuffer.Width, _serverBuffer.Height));
                        _serverBuffer.DrainDirtyRects();
                        _frameStarted = true;
                        _isFirstFrame = true;

                        await SendResize();
                    }
                    else if (_clientPendingBuffer.HasDirtyRects)
                    {
                        // We haven't taken an action yet in this pass, and
                        // there are dirty rects to send, so send the next one
                        // to the client.
                        Int32Rect dirtyRect = _clientPendingBuffer.DrainDirtyRect();

                        if (_isFirstFrame)
                        {
                            // The first frame is sent as a regular image,
                            // that is used by the client to replace the
                            // contents of their buffer.
                            _isFirstFrame = false;
                            _clientBuffer.SetBits(_clientPendingBuffer.View, dirtyRect);
                            await SendDrawImage(_clientBuffer.View, dirtyRect, true);
                        }
                        else
                        {
                            // All other dirty rects are sent as "delta"
                            // images which are encoded via XOR so that only
                            // the bits that are different are sent as "1".
                            // This should result in savings where the dirty
                            // rect actually contains a significant number of
                            // unchanged bits because the "0" bits should
                            // compress well.

                            // TODO: XOR
                            //uint[] clientContent = _clientBuffer.GetBits(dirtyRect);
                            _clientBuffer.SetBits(_clientPendingBuffer.View, dirtyRect);

                            //for (int i = 0; i < content.Length; i++)
                            //{
                            //    content[i] ^= clientContent[i];
                            //}

                            await SendDrawImage(_clientBuffer.View, dirtyRect, false);
                        }
                    }
                    else
                    {
                        // The client pending buffer is clean, meaning that we
                        // have already sent any deltas to the client.
                        if (_frameStarted)
                        {
                            _frameStarted = false;

                            // We have finished sending all deltas to the
                            // client, so we send a present.
                            await SendPresent();
                        }
                        else if (_clientPendingBuffer.Width != _serverBuffer.Width || _clientPendingBuffer.Height != _serverBuffer.Height)
                        {
                            // The server buffer is a different size.  We need to
                            // reset the client buffers.
                            _clientPendingBuffer = null;
                            _clientBuffer = null;
                        }
                        else if (_serverBuffer.ContentsAreComplete && _serverBuffer.HasDirtyRects)
                        {
                            // The server buffer is the same size as before but
                            // has accumulated additional changes.  We drain
                            // dirty rects into our pending buffer which we will
                            // eventually send to the client.  This group of
                            // dirty rects will constitute a "frame" to present.
                            foreach (Int32Rect dirtyRect in _serverBuffer.DrainDirtyRects())
                            {
                                _clientPendingBuffer.SetBits(_serverBuffer.View, dirtyRect);
                            }

                            _frameStarted = true;
                        }
                        else
                        {
                            // The server buffer is either clean, or is in the
                            // middle of collecting new deltas, meaning that no
                            // changes are ready to send to the client.  There
                            // is nothing more to do.
                            actionTaken = false;
                        }
                    }

                } while (actionTaken && _pendingAckQueue.Count < MAX_PENDING_ACK); // There may be more to send, so loop
            }
            finally
            {
                _sending = false;
            }
        }

        private async Task SendResize()
        {
            int messageId = _ackRandom.Next();
            _pendingAckQueue.Enqueue(messageId);

            var packet = new
            {
                Type = "Resize",
                Width = _clientBuffer.Width,
                Height = _clientBuffer.Height,
                MessageId = messageId
            };

            byte[] message = Encoding.ASCII.GetBytes(_serializer.Serialize(packet));

            if (_webSocket.State == WebSocketState.Open)
            {
                await _webSocket.SendAsync(new ArraySegment<byte>(message), WebSocketMessageType.Text, true, CancellationToken.None);
            }
        }

        private async Task SendDrawImage(Buffer2DView<uint> srcView, Int32Rect rect, bool mode)
        {
            int messageId = _ackRandom.Next();
            _pendingAckQueue.Enqueue(messageId);

            // Create a bitmap for the contents of each rect.
            BitmapSource bitmap = null;
            Buffer2DView<uint> rectView = new Buffer2DView<uint>(srcView, rect);

            if (_encoderType == 0 && _qualityLevel == 1)
            {
                // send an xor-encoded PNG
                WriteableBitmapBuffer xorBitmapBuffer = new WriteableBitmapBuffer(rectView);
                xorBitmapBuffer.XorBits(_clientBuffer.View, rect, new Int32Point(0, 0));
                bitmap = xorBitmapBuffer.View.CreateBitmapSource(96.0, 96.0, PixelFormats.Pbgra32, null);
            }
            if (_encoderType == 0 && _qualityLevel == 2)
            {
                // send an empty PNG
                BitmapBuffer emptyBuffer = new BitmapBuffer(rect.Width, rect.Height);
                bitmap = emptyBuffer.View.CreateBitmapSource(96.0, 96.0, PixelFormats.Pbgra32, null);
            }
            else
            {
                // send a full-color PNG or JPG
                bitmap = rectView.CreateBitmapSource(96.0, 96.0, PixelFormats.Pbgra32, null);
            }

            MemoryStream memoryStream = new MemoryStream();

            if (_encoderType == 0)
            {
                // Encode the dirty pixels into a PNG.
                PngBitmapEncoder pngEncoder = new PngBitmapEncoder();
                pngEncoder.Frames.Add(BitmapFrame.Create(bitmap));
                pngEncoder.Save(memoryStream);
            }
            else
            {
                JpegBitmapEncoder jpgEncoder = new JpegBitmapEncoder();
                jpgEncoder.QualityLevel = _qualityLevel;
                jpgEncoder.Frames.Add(BitmapFrame.Create(bitmap));
                jpgEncoder.Save(memoryStream);
            }

            // Convert to data-url
            byte[] pngData = memoryStream.ToArray();
            string imgDataUrl = "data:image/png;base64," + System.Convert.ToBase64String(pngData);

            var packet = new
            {
                Type = "DrawImage",
                X = rect.X,
                Y = rect.Y,
                MessageId = messageId,
                DataUrl = imgDataUrl
            };

            byte[] message = Encoding.ASCII.GetBytes(_serializer.Serialize(packet));

            if (_webSocket.State == WebSocketState.Open)
            {
                await _webSocket.SendAsync(new ArraySegment<byte>(message), WebSocketMessageType.Text, true, CancellationToken.None);
            }
        }

        private async Task SendPresent()
        {
            int messageId = _ackRandom.Next();
            _pendingAckQueue.Enqueue(messageId);

            var packet = new
            {
                Type = "Present",
                MessageId = messageId,
            };

            byte[] message = Encoding.ASCII.GetBytes(_serializer.Serialize(packet));

            if (_webSocket.State == WebSocketState.Open)
            {
                await _webSocket.SendAsync(new ArraySegment<byte>(message), WebSocketMessageType.Text, true, CancellationToken.None);
            }
        }

        private async void ScreenScraper_Updated(object sender, BitmapBufferUpdatedEventArgs e)
        {
            _serverBuffer.ContentsAreComplete = e.ContentsAreComplete;

            // Copy the update into the server buffer.
            _serverBuffer.SetBits(e.View, e.Rect);

            if (!_sending && _pendingAckQueue.Count < MAX_PENDING_ACK)
            {
                await SendNextClientMessage();
            }
        }

        private async void ScreenScraper_Resized(object sender, EventArgs e)
        {
            // Replace server back buffer completely.
            _serverBuffer = new DirtyRectBitmapBuffer(_screenScraper.Width, _screenScraper.Height, MAX_DIRTYRECTS);

            if (!_sending && _pendingAckQueue.Count < MAX_PENDING_ACK)
            {
                await SendNextClientMessage();
            }
        }

        private int ToInt(object o)
        {
            if (o is int)
            {
                return (int)o;
            }
            else if (o is decimal)
            {
                return Decimal.ToInt32((decimal)o);
            }
            else if (o is float)
            {
                return (int)(float)o;
            }
            else if (o is double)
            {
                return (int)(double)o;
            }
            else
            {
                throw new InvalidCastException("unable to cast object to int");
            }
        }

        private const int MAX_DIRTYRECTS = 8;
        private const int MAX_PENDING_ACK = 10;

        private WebSocket _webSocket;
        private ScreenScraperBitmapBuffer _screenScraper;
        private DirtyRectBitmapBuffer _serverBuffer;
        private DirtyRectBitmapBuffer _clientPendingBuffer;
        private WriteableBitmapBuffer _clientBuffer;
        private CancellationTokenSource _cancellationToken;
        private Random _ackRandom;
        private Queue<int> _pendingAckQueue = new Queue<int>(MAX_PENDING_ACK);
        private bool _sending;
        private bool _frameStarted;
        private bool _isFirstFrame;
        private int _encoderType; // 0 = png, 1 = jpg
        private int _qualityLevel; // 1..100 for jpg
        private JavaScriptSerializer _serializer;
    }
}
