﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using Microsoft.DwayneNeed.Extensions;
using Microsoft.DwayneNeed.Utilities;
using Microsoft.DwayneNeed.Numerics;

namespace RcpServer
{
    public class DirtyRectBitmapBuffer : WriteableBitmapBuffer
    {
        public DirtyRectBitmapBuffer(int width, int height, int maxDirtyRects)
            : base(width, height)
        {
            _dirtyRects = new Int32Rect[maxDirtyRects];
            for (int i = 0; i < maxDirtyRects; i++)
            {
                _dirtyRects[i] = Int32Rect.Empty;
            }

            // Start with one dirty rect.
            _dirtyRects[0] = new Int32Rect(0, 0, width, height);
        }

        // This is a hack to pass along a hint of when the partial updates to
        // the buffer are complete.  This will toggle between true and false as
        // updates are processed.  Not every frame will make it to the client.
        public bool ContentsAreComplete { get; set; }

        public override void SetBits(Buffer2DView<uint> srcView, Int32Rect srcRect, Int32Point? dstPoint = null)
        {
            base.SetBits(srcView, srcRect, dstPoint);
            MergeDirtyRect(srcRect);
        }

        public bool HasDirtyRects
        {
            get
            {
                for (int i = 0; i < _dirtyRects.Length; i++)
                {
                    Int32Rect dirtyRect = _dirtyRects[i];
                    if (!dirtyRect.IsEmpty)
                    {
                        return true;
                    }
                }

                return false;
            }
        }

        public Int32Rect DrainDirtyRect()
        {
            for (int i = 0; i < _dirtyRects.Length; i++)
            {
                Int32Rect dirtyRect = _dirtyRects[i];
                if (!dirtyRect.IsEmpty)
                {
                    _dirtyRects[i] = Int32Rect.Empty;
                    return dirtyRect;
                }
            }

            return Int32Rect.Empty;
        }

        public IEnumerable<Int32Rect> DrainDirtyRects()
        {
            List<Int32Rect> dirtyRects = new List<Int32Rect>();

            for (int i = 0; i < _dirtyRects.Length; i++)
            {
                Int32Rect dirtyRect = _dirtyRects[i];
                if (!dirtyRect.IsEmpty)
                {
                    _dirtyRects[i] = Int32Rect.Empty;
                    dirtyRects.Add(dirtyRect);
                }
            }

            return dirtyRects;
        }

        private void MergeDirtyRect(Int32Rect dirtyRect)
        {
            // If we are not tracking dirty rects, return immediately.
            if(_dirtyRects.Length == 0)
            {
                return;
            }

            // First, look for an empty rect to replace.
            bool dirtyRectAdded = false;
            for (int i = 0; i < _dirtyRects.Length; i++)
            {
                if (_dirtyRects[i].IsEmpty)
                {
                    _dirtyRects[i] = dirtyRect;
                    dirtyRectAdded = true;
                    break;
                }
            }

            // If needed, search for the least expensive merge.
            if(!dirtyRectAdded)
            {
                if(_dirtyRects.Length == 1)
                {
                    _dirtyRects[0] = dirtyRect;
                }

                int iLeastAdditionalArea = 0;
                int leastAdditionalArea = _dirtyRects[0].Union(dirtyRect).GetArea() - _dirtyRects[0].GetArea();
                for(int i = 1; i < _dirtyRects.Length; i++)
                {
                    int testArea = _dirtyRects[i].Union(dirtyRect).GetArea() - _dirtyRects[i].GetArea();
                    if(testArea < leastAdditionalArea)
                    {
                        leastAdditionalArea = testArea;
                        iLeastAdditionalArea = i;
                    }
                }

                _dirtyRects[iLeastAdditionalArea] = _dirtyRects[iLeastAdditionalArea].Union(dirtyRect);
                dirtyRectAdded = true;
            }
        }

        private Int32Rect[] _dirtyRects;
    }
}
