﻿module RCP {
    // http://dev.w3.org/html5/websockets/
    export interface IWebSocketModel {
        close(code?: number, reason?: string): void;
        send(data: string): void;
        onmessage : (message:string)=>void;
        bandwidth(value? : number) : number; // mega bits per second
        latency(value? : number) : number; // in seconds
        readyState(): number;
    }

    // The NetworkSegmentMessage class tracks a message being transmitted
    // through the network segment.  The message itself is a simple string.
    // The time is stored for when the message will have been fully transmitted
    // through the network segment and available at the other end.
    class NetworkSegmentMessage {
        public message: string;
        public availableTime: number;
    }

    // The NetworkSegment class simulates a network segment with a certain
    // bandwidth and latency characteristics.  Messages can be sent through
    // this simulated network and will arrive at the other end with the
    // appropriate delay.
    class NetworkSegment {
        constructor() {
            this._pendingMessages = new NetworkSegmentMessage[];
            this._nextWriteTime = Date.now();
            this._bandwidth = 0;
            this._latency = 0;
            this._timer = 0;
        }

        // The bandwidth of the network segment, in megabits per second.
        public bandwidth(value?: number): number {
            if (value != undefined) {
                this._bandwidth = value;
            }
            return this._bandwidth;
        }

        // The latency of the network segment, in seconds.
        public latency(value?: number): number {
            if (value != undefined) {
                this._latency = value;
            }
            return this._latency;
        }

        // Queue a message to be sent through the network segment.
        public send(message: string): void {
            var now = Date.now();
            var messageStartTime = now;
            if (this._nextWriteTime > now) {
                messageStartTime = this._nextWriteTime;
            }

            var messageEndTime = messageStartTime;
            if (this._bandwidth != 0) {
                messageEndTime += (message.length * 8 / (this._bandwidth * 1024 * 1024)) * 1000;
            }
            this._nextWriteTime = messageEndTime;

            var messageAvailableTime = messageEndTime;
            if (this._latency != 0) {
                messageAvailableTime += this._latency * 1000;
            }

            var segmentMessage = new NetworkSegmentMessage();
            segmentMessage.message = message;
            segmentMessage.availableTime = messageAvailableTime;

            this._pendingMessages.push(segmentMessage);
            this._processPendingMessages();
        }

        // Listen for messages arriving at the end of the network segment.
        public onmessage : (message: string) => void;

        // Dequeue any pending messages that are now available, and raise the
        // onmessage event for each.
        private _processPendingMessages(): void {
            var me = this;

            if (this._timer != 0) {
                clearTimeout(this._timer);
                this._timer = 0;
            }

            while (this._pendingMessages.length > 0) {
                var message = this._pendingMessages[0];
                var now = Date.now();

                if (message.availableTime <= now) {
                    this._pendingMessages.shift();

                    if (this.onmessage != null) {
                        this.onmessage(message.message);
                    }
                } else {
                    break;
                }
            }

            if (this._pendingMessages.length > 0) {
                this._timer = setTimeout(function () {
                    me._timer = 0;
                    me._processPendingMessages();
                }, this._pendingMessages[0].availableTime - Date.now());
            }
        }

        private _bandwidth: number;
        private _latency: number;
        private _pendingMessages: NetworkSegmentMessage[];
        private _nextWriteTime: number;
        private _timer: number;
    }

    export class WebSocketModel implements IWebSocketModel {
        constructor(endpoint: string) {
            var me = this;
            var now = new Date();

            // Create the network segment for reading messages from the
            // websocket.  When a message arrives, raise the onmessage event.
            this._readNetworkSegment = new NetworkSegment();
            this._readNetworkSegment.onmessage = function(message: string): void {
                if (me.onmessage != null) {
                    me.onmessage(message);
                }
            }

            // Create the network segment for writing messages to the
            // websocket.  When a message arrives, send it to the websocket.
            this._writeNetworkSegment = new NetworkSegment();
            this._writeNetworkSegment.onmessage = function (message: string): void {
                if (me._webSocket != null) {
                    me._webSocket.send(message);
                }
            }

            // Create the websocket itself.  When a message arrives, send it
            // to the network segment for reading from the websocket.
            this._webSocket = new WebSocket(endpoint);
            this._webSocket.onmessage = function (args) {
                me._readNetworkSegment.send(args.data);
            }
        }
        
        public close(code?: number, reason?: string): void {
            if (this._webSocket != null) {
                this._webSocket.close(code, reason);
            } else {
                throw new Error("WebSocket is not available.");
            }
        }

        public send(message: string): void {
            // Send the message to the network segment for writing to the
            // websocket.
            this._writeNetworkSegment.send(message);
        }

        public onmessage : (message: string) => void;

        public readyState(): number {
            return this._webSocket.readyState;
        }

        // The bandwidth of the additional network segments in the websocket
        // model, in megabits per second.
        public bandwidth(value?: number): number {
            this._readNetworkSegment.bandwidth(value);
            return this._writeNetworkSegment.bandwidth(value);
        }

        // The latency of the additional network segments in the websocket
        // model, in seconds.
        public latency(value?: number): number {
            this._readNetworkSegment.latency(value);
            return this._writeNetworkSegment.latency(value);
        }

        private _webSocket : WebSocket;
        private _readNetworkSegment: NetworkSegment;
        private _writeNetworkSegment: NetworkSegment;
    }
}