/// <reference path="WebSocketModel.ts"/>
/// <reference path="jquery.d.ts"/>
/// <reference path="perfstore.ts"/>
/// <reference path="perfmon.ts"/>

interface IPacket {
    Type: string;
    MessageId: number;
}

class Packet implements IPacket {
    Type: string;
    MessageId: number;
}

class DrawImage {
    X: number;
    Y: number;
    Image: HTMLImageElement;
}

class Frame {
    Width: number;
    Height: number;
    PendingImageCount: number;
    Present: bool;
    Images: DrawImage[];
}

interface IResizePacket extends IPacket {
    Width: number;
    Height: number;
}

interface IDrawImagePacket extends IPacket {
    X: number;
    Y: number;
    DataUrl: string;
}

interface IPresentPacket extends IPacket {
}

class MouseMovePacket extends Packet {
    X: number;
    Y: number;
}

class MouseButtonPacket extends Packet {
    Button: number;
}

class KeyboardPacket extends Packet {
    KeyCode: number;
}

class QOSPacket extends Packet {
    Format: string;
    QualityLevel: number;
}

var _batchFrameUpdates: bool;
var _showPixels: bool;
var _showDirtyRects: bool;
var _showStats: bool;

var _pixelCanvas: HTMLCanvasElement;
var _dirtyRectCanvas: HTMLCanvasElement;
var _statsCanvas: HTMLCanvasElement;

var _webSocket: RCP.IWebSocketModel = null;
var _frames = new Frame[];

var _perfStore: Telemetry.PerfStore;
var _perfmon: PerfMonitor;

function connect() {
    var connectionString = <HTMLInputElement> document.getElementById('connectionString');
    _webSocket = new RCP.WebSocketModel(connectionString.value);
    _webSocket.latency(parseInt((<HTMLInputElement> document.getElementById("AdditionalLatency")).value)/1000);
    _webSocket.bandwidth(parseFloat((<HTMLInputElement> document.getElementById("AdditionalBandwidth")).value));
    _webSocket.onmessage = processMessage;
}

function resetCounters() {
    _perfStore.Reset(); 
    _perfmon.Reset();
}

function processMessage(message: string) {
    try {
        var messageId = onMessage(message);

        var packet = new Packet();
        packet.Type = "Ack";
        packet.MessageId = messageId;
        _webSocket.send(JSON.stringify(packet));
    }
    catch (e) {
        _webSocket.close();
    }
}

function onMessage(message: string): number {
    var packet = <IPacket> JSON.parse(message);

    switch (packet.Type) {
        case "Resize":
            onResize((<IResizePacket>packet).Width, (<IResizePacket>packet).Height);
            break;
        case "DrawImage":
            onDrawImage((<IDrawImagePacket>packet).X, (<IDrawImagePacket>packet).Y, (<IDrawImagePacket>packet).DataUrl);
            break;
        case "Present":
            onPresent();
            break;
        default:
            throw new Error("Unrecognized packet type.");
    }

    return packet.MessageId;
}

function onResize(width: number, height: number) {

    var frame = new Frame();
    frame.Width = width;
    frame.Height = height;
    frame.Present = false;
    frame.PendingImageCount = 0;
    frame.Images = new DrawImage[];

    _frames.push(frame);
}

function onDrawImage(x: number, y: number, imageData: string) {
    var perfkey = _perfStore.Start("DrawImage");
    // If batching frame updates, append this DrawImage to the last frame.
    // If the last frame is already marked to be presented, or if there
    // are no frames at all, create a new frame.
    //
    // If not batching frame updates, always create a new frame for each
    // DrawImage.
    var frame: Frame; 
    
    if (_frames.length == 0 || _frames[_frames.length - 1].Present || !_batchFrameUpdates) {
        frame = new Frame();
        frame.Present = !_batchFrameUpdates;
        frame.Width = NaN;
        frame.Height = NaN;
        frame.Images = new DrawImage[];
        frame.PendingImageCount = 0;

        _frames.push(frame);
    } else {
        frame = _frames[_frames.length - 1];
    }

    // Create an image for the image data, and hook up the onLoad callback.
    var image = new Image;
    image.dataFld = perfkey.toString();
    image.onload = function () {
        frame.PendingImageCount--;
        var perfkey = this.dataFld;
        _perfStore.End("DrawImage", perfkey, (imageData.length / 1000000) * 8);
        // Any time an image finishes loading, try to present any pending frames.
        presentFrames();
    }

    var drawImage = new DrawImage();
    drawImage.X = x;
    drawImage.Y = y;
    drawImage.Image = image;

    frame.Images.push(drawImage);
    frame.PendingImageCount++;

    // Start loading the image data.
    image.src = imageData;
}

function onPresent() {
    if (_batchFrameUpdates && _frames.length > 0) {
        // Mark the current frame as ready to present.
        var frame = _frames[_frames.length - 1];
        frame.Present = true;

        presentFrames();
    }
}

function presentFrames() {
    
    // Starting from the front, present any frames that are ready.
    // We cannot present frames out of order, and we cannot skip
    // frames.
    while(_frames.length > 0 && _frames[0].Present && _frames[0].PendingImageCount == 0) {
        presentFrame(_frames.shift());
    }
    
}

function presentFrame(frame: Frame) {
    var perfkey = _perfStore.Start("PresentFrame");
    // Update the width of the front buffers if needed.
    if (!isNaN(frame.Width) && frame.Width != _pixelCanvas.width) {
        _pixelCanvas.width = frame.Width;
        _dirtyRectCanvas.width = frame.Width;
    }

    // Update the height of the front buffers if needed.
    if (!isNaN(frame.Height) && frame.Height != _pixelCanvas.height) {
        _pixelCanvas.height = frame.Height;
        _dirtyRectCanvas.height = frame.Height;
    }

    var pixelContext = _pixelCanvas.getContext("2d");
    var dirtyRectContext = _dirtyRectCanvas.getContext("2d");
    var statsContext = _statsCanvas.getContext("2d");

    // Clear the dirty rect canvas, we don't want all of the dirty rects
    // to just pile up.
    dirtyRectContext.clearRect(0, 0, _dirtyRectCanvas.width, _dirtyRectCanvas.height);

    frame.Images.forEach(function (value: DrawImage, index: number, array: DrawImage[]) {
        // The image data is available, so just draw the image.  The front
        // buffer global composite operation is expected to be source-over.
        pixelContext.drawImage(value.Image, value.X, value.Y);

        // Draw a filled, partially transparent rectangle the size of the
        // image to the dirty rect buffer.  This helps visualize what parts
        // of the frame are changing, as well as to see a heat map of
        // overlapping dirty rects.
        dirtyRectContext.globalAlpha = 0.50;
        dirtyRectContext.fillStyle = "#ff0000";
        dirtyRectContext.fillRect(value.X, value.Y, value.Image.width, value.Image.height);
    });

    // Clear the stats canvas, we don't want the text to pile up.
    statsContext.clearRect(0, 0, _statsCanvas.width, _statsCanvas.height);
    statsContext.font = "20px sans-serif";
    statsContext.fillText("Updates Per Frame: " + frame.Images.length, 1, 20);
    _perfStore.End("PresentFrame", perfkey, 1);
}

function onBatchFrameUpdatesChanged() {
    _batchFrameUpdates  = (<HTMLInputElement> document.getElementById("BatchFrameUpdates")).checked;

    // If no longer batching frame updates, we need to mark all existing
    // frames as ready to present.
    _frames.forEach(function (value: Frame, index: number, array: Frame[]) {
        value.Present = true;
    });
}

function onShowPixelsChanged() {
    _showPixels = (<HTMLInputElement> document.getElementById("ShowPixels")).checked;
    if (_showPixels) {
        _pixelCanvas.style.visibility = "visible";
    } else {
        _pixelCanvas.style.visibility = "hidden";
    }
}

function onShowDirtyRectsChanged() {
    _showDirtyRects= (<HTMLInputElement> document.getElementById("ShowDirtyRects")).checked;
    if (_showDirtyRects) {
        _dirtyRectCanvas.style.visibility = "visible";
    } else {
        _dirtyRectCanvas.style.visibility = "hidden";
    }
}

function onShowStatsChanged() {
    _showStats = (<HTMLInputElement> document.getElementById("ShowStats")).checked;
    if (_showStats) {
        _statsCanvas.style.visibility = "visible";
    } else {
        _statsCanvas.style.visibility = "hidden";
    }
}

function onAdditionalLatencyChanged() {
    if (_webSocket != null) {
        _webSocket.latency(parseInt((<HTMLInputElement> document.getElementById("AdditionalLatency")).value)/1000);
    }
}

function onAdditionalBandwidthChanged() {
    if (_webSocket != null) {
        _webSocket.bandwidth(parseFloat((<HTMLInputElement> document.getElementById("AdditionalBandwidth")).value));
    }
}

function onMouseMove(event: MouseEvent) {
    if (_webSocket != null && _webSocket.readyState() == 1) {
        // Get the offsets for the pixel canvas.
        var element: HTMLElement;
        var offsetX = 0;
        var offsetY = 0;
        for (element = _pixelCanvas; element != null; element = <HTMLElement> element.offsetParent) {
            offsetX += element.offsetLeft;
            offsetY += element.offsetTop;
        }

        var packet = new MouseMovePacket();
        packet.Type = "MouseMove";
        packet.MessageId = 0; // Not used in messages sent to the server
        packet.X = event.pageX - offsetX;
        packet.Y = event.pageY - offsetY;

        _webSocket.send(JSON.stringify(packet));
    }

    event.cancelBubble = true;
    if (event.stopPropagation) {
        event.stopPropagation();
        event.preventDefault();
    }
}

function onMouseDown(event: MouseEvent) {
    if (_webSocket != null && _webSocket.readyState() == 1) {
        var packet = new MouseButtonPacket();
        packet.Type = "MouseDown";
        packet.MessageId = 0; // Not used in messages sent to the server
        packet.Button = event.button;

        _webSocket.send(JSON.stringify(packet));
    }

    _pixelCanvas.focus();

    event.cancelBubble = true;
    if (event.stopPropagation) {
        event.stopPropagation();
        event.preventDefault();
    }
}

function onMouseUp(event: MouseEvent) {
    if (_webSocket != null && _webSocket.readyState() == 1) {
        var packet = new MouseButtonPacket();
        packet.Type = "MouseUp";
        packet.MessageId = 0; // Not used in messages sent to the server
        packet.Button = event.button;

        _webSocket.send(JSON.stringify(packet));
    }

    event.cancelBubble = true;
    if (event.stopPropagation) {
        event.stopPropagation();
        event.preventDefault();
    }
}

function onKeyDown(event: KeyboardEvent) {
    if (_webSocket != null && _webSocket.readyState() == 1) {
        var packet = new KeyboardPacket();
        packet.Type = "KeyDown";
        packet.MessageId = 0; // Not used in messages sent to the server
        packet.KeyCode = event.keyCode;

        _webSocket.send(JSON.stringify(packet));
    }

    event.cancelBubble = true;
    if (event.stopPropagation) {
        event.stopPropagation();
        event.preventDefault();
    }
}

function onKeyUp(event: KeyboardEvent) {
    if (_webSocket != null && _webSocket.readyState() == 1) {
        var packet = new KeyboardPacket();
        packet.Type = "KeyUp";
        packet.MessageId = 0; // Not used in messages sent to the server
        packet.KeyCode = event.keyCode;

        _webSocket.send(JSON.stringify(packet));
    }

    event.cancelBubble = true;
    if (event.stopPropagation) {
        event.stopPropagation();
        event.preventDefault();
    }
}

function onContextMenu(event: MouseEvent) {
    return false;
}

function onQualityChanged(value: number) {
    if (_webSocket != null && _webSocket.readyState() == 1) {
        var packet = new QOSPacket();
        packet.Type = "QOS";
        packet.MessageId = 0; // Not used in messages sent to the server

        switch (value) {
            case "0": packet.Format = "png"; packet.QualityLevel = 0; break;
            case "1": packet.Format = "png"; packet.QualityLevel = 1; break;
            case "2": packet.Format = "png"; packet.QualityLevel = 2; break;
            case "3": packet.Format = "jpg"; packet.QualityLevel = 100; break;
            case "4": packet.Format = "jpg"; packet.QualityLevel = 80; break;
            case "5": packet.Format = "jpg"; packet.QualityLevel = 60; break;
            case "6": packet.Format = "jpg"; packet.QualityLevel = 40; break;
            case "7": packet.Format = "jpg"; packet.QualityLevel = 20; break;
            case "8": packet.Format = "jpg"; packet.QualityLevel = 1; break;
        }

        _webSocket.send(JSON.stringify(packet));
    }
}

window.onload = () => {
      // Perf Counters

    _perfStore = Telemetry.PerfStore.getInstance();
    _perfmon = new PerfMonitor($("#perfmoncanvas")[0]);
    _perfmon.start();

    _batchFrameUpdates  = (<HTMLInputElement> document.getElementById("BatchFrameUpdates")).checked;
    
    _pixelCanvas = <HTMLCanvasElement> document.getElementById("pixelCanvas");
    _dirtyRectCanvas = <HTMLCanvasElement> document.getElementById("dirtyRectCanvas");
    _statsCanvas = <HTMLCanvasElement> document.getElementById("statsCanvas");

    _pixelCanvas.onmousemove = onMouseMove;
    _pixelCanvas.onmousedown = onMouseDown;
    _pixelCanvas.onmouseup = onMouseUp;
    _pixelCanvas.onkeydown = onKeyDown;
    _pixelCanvas.onkeyup = onkeyup;
    _pixelCanvas.setAttribute("tabindex", "0"); // HACK: make the canvas focusable
    _pixelCanvas.contentEditable = "true"; // HACK: make the canvas focusable
    _pixelCanvas.oncontextmenu = onContextMenu; // HACK: disable the browser context menu

    // HACK: hook the same events to the layers about the canvas, since there is no "click through" in IE.
    _dirtyRectCanvas.onmousemove = onMouseMove;
    _dirtyRectCanvas.onmousedown = onMouseDown;
    _dirtyRectCanvas.onmouseup = onMouseUp;
    _dirtyRectCanvas.onkeydown = onKeyDown;
    _dirtyRectCanvas.onkeyup = onkeyup;
    _dirtyRectCanvas.oncontextmenu = onContextMenu; // HACK: disable the browser context menu

    // HACK: hook the same events to the layers about the canvas, since there is no "click through" in IE.
    _statsCanvas.onmousemove = onMouseMove;
    _statsCanvas.onmousedown = onMouseDown;
    _statsCanvas.onmouseup = onMouseUp;
    _statsCanvas.onkeydown = onKeyDown;
    _statsCanvas.onkeyup = onkeyup;
    _statsCanvas.oncontextmenu = onContextMenu; // HACK: disable the browser context menu

  

};
